/****************************************************************************
  Module : CovGraph.cpp
  Contents : Method definitions for the grphWin and GraphPane classes.
  These functions control the display and behavior of a graph window.
****************************************************************************/
#include <stdio.h> //For sprintf
#include <math.h> //For sqrt
#include <v/vbrush.h>
#include <v/vcolor.h>
#include <v/vmenu.h>
#include <v/vfont.h>

#include "robapp.h"
#include "menus.h"
#include "proto.h"
#include "covgraph.h"
#include "platform.h"

/****************************************************************************
  Constructor: grphWin
  Parameters: name, the title of the window
  h, the height of the window
  w, the width of the window
  cd, the data set to be displayed by the window
****************************************************************************/
covgrphWin::covgrphWin(void)
	: vCmdWindow("Graph Display",700,400)
{ 
  mp = new vMenuPane(MainMenu);
  AddPane(mp);
  gp = new CovGraphPane();
  AddPane(gp);
  
  ShowWindow();
  gp->ShowHScroll(1);
  gp->SetHScroll(10,50);
  
  //Set font
  //vFont myfont(vfFixed);
  //gp->SetFont(myfont);
  gp->Resize(0,0);
  
}

/*****************************************************************************
  Destructor : covgrphWin
*****************************************************************************/

covgrphWin::~covgrphWin(void)
{
  delete mp;
  delete gp;
}

void covgrphWin::CloseWin(void)
{
  myApp->CloseNotify(this);
  vCmdWindow::CloseWin();
}

void covgrphWin::SetDataSet(int x)
{
  gp->SetDataSet(x);
  if (myApp->Params.title[0] != '\0') {
    char title[40];
    sprintf(title,"Graph Display - %s",myApp->Params.title);
    SetTitle(title);
  }
  ShowWindow();
}

/****************************************************************************
  Constructor: CovGraphPane
  Parameters: cd, data set to be displayed.
****************************************************************************/

CovGraphPane::CovGraphPane(void)
	: vCanvasPane()
{
  //Setup values for displaying the window
  LeftMargin = 90; //Position of the left hand part of the graph
  TopMargin = 10; //Position of top of graph
  FontHeight = 12;
  FontWidth = 6;
  Yextent = 1;
  Xextent = 1;
  Mark = 0.0;
}

/***************************************************************************
  Method: Resize
  Parameters: newH, the new height of the window
      	      newW, the new width of the window
  Purpose: Resize is called whenever the size of the window changes. It
   	   recalculates parameters for the graph. If it is called with
           newH == 0, it determines the height and width to use by itself.
           This kludge is used to properly set height and width before the
           first time the graph is displayed and after the canvas has been
           added to the window (Trust me, it has to work this way).
****************************************************************************/

void CovGraphPane::Resize(int newW, int newH)
{
  if (!newH) {
    Height = GetHeight();
    Width = GetWidth();
  }
  else {
    Height = newH;
    Width = newW;
  }

  BottomMargin = Height - 55;
  RightMargin = Width - 10;
  Xint = (RightMargin - LeftMargin) / Xextent;
  Yint = (BottomMargin - TopMargin) / Yextent;

  #ifdef __UNIX__
  if (newH) Redraw(0,0,0,0);
  #endif
}

/*****************************************************************************
    Method: SetDataSet
    Parameters: dg, 1 for orig, 2 for model
*****************************************************************************/
void CovGraphPane::SetDataSet(int dg)
{
  CoverageDataSet* cd1 = myApp->CovData;
  CoverageDataSet* cd2 = NULL;
  if (dg == 2) cd2 = myApp->CovModData;

  Xextent = 1.0;

  if (cd2)
    Yextent = max(cd1->cov_data[cd1->Size - 1].faults,
		  cd2->cov_data[cd2->Size - 1].faults);
  else
    Yextent = cd1->cov_data[cd1->Size - 1].faults;

   Xint = (RightMargin - LeftMargin) / Xextent;
   Yint = (BottomMargin - TopMargin) / Yextent;
   data = dg;
   Redraw(0,0,0,0);
}

/****************************************************************************
  Method : Redraw
  Parameters : sx, the left most column that needs to be redrawn
  sy, the top most row that needs to be redrawn
  height, the height of the area to be redrawn
  width, the width of the area to be redrawn
  Purpose : Redraw is called whenever the graph pane has to be redrawn
****************************************************************************/

void CovGraphPane::Redraw(int, int, int, int)
{
  //temp variables
  int x,y;
  int len;
  char str[64];
  //SetBackground(vStdColors[vC_White]);
  Clear();
  vBrush mybrush;
  mybrush.SetColor(vStdColors[vC_MedGray]);
  SetBrush(mybrush);
  DrawRectangle(LeftMargin+2,TopMargin,RightMargin-LeftMargin-2,
		BottomMargin-TopMargin-2);

  /*
  y = Height - 5;
  x = LeftMargin + 2;
  len = sprintf(str,"Alpha 0: %9e",myApp->Params.alpha0);
  DrawText(x,y,str);
  x += (len + 3) * FontWidth;
  len = sprintf(str,"Alpha 1: %9e",myApp->Params.alpha1);
  DrawText(x,y,str);
  x += (len + 3) * FontWidth;
  sprintf(str,"Alpha 2: %9e",myApp->Params.alpha2);
  DrawText(x,y,str);
  */
  //Print out the left hand label
  vFont myfont(vfFixed,12);
  SetFont(myfont);

  vPen mypen(0,128,0);
  SetPen(mypen);
  y = (BottomMargin - TopMargin) / 2 + TopMargin;
  DrawText(2,y - (FontHeight + 3) * 3,"  F");
  DrawText(2,y - (FontHeight + 3) * 2,"T A");
  DrawText(2,y - (FontHeight + 3),"O U");
  DrawText(2,y,"T L");
  DrawText(2,y + (FontHeight + 3),"A T");
  DrawText(2,y + (FontHeight + 3) * 2,"L S");
  //Print out bottom label
  x = (((Width - LeftMargin) / 2) - (FontWidth * 4)) + LeftMargin;
  y = BottomMargin + (FontHeight << 1)+3;
  DrawText(x,y,"Coverage");
  myfont.SetFontValues(vfSerif,10);
  SetFont(myfont);

  //Print out the bottom scale
  myfont.SetFontValues(vfSerif,12);
  SetFont(myfont);
  x = LeftMargin - FontWidth * 5 + 5;
  y = BottomMargin + FontHeight;
  DrawText(x,y,"0.00");

  x = (RightMargin - LeftMargin) / 4 + LeftMargin+1;
  len = sprintf(str,"25");
  DrawText(x - (FontWidth * (len / 2)),y+2,str);
  DrawLine(x,BottomMargin - 2,x,BottomMargin + 4); //Tick line

  x = (RightMargin - LeftMargin) * 1/2 + LeftMargin+3;
  len = sprintf(str,"50");
  DrawText(x - (FontWidth * (len / 2)),y+2,str);
  DrawLine(x,BottomMargin - 2,x,BottomMargin + 4); //Tick line

  x = (RightMargin - LeftMargin) * 3/4 + LeftMargin+4;
  len = sprintf(str,"75");
  DrawText(x - (FontWidth * (len / 2)),y+2,str);
  DrawLine(x,BottomMargin - 2,x,BottomMargin + 4); //Tick line

  x = RightMargin;
  len = sprintf(str,"100");
  DrawText(x - FontWidth * len+3,y+2,str);
  DrawLine(x,BottomMargin - 2,x,BottomMargin + 4); //Tick line

  //Print left hand scale
  y = (BottomMargin - TopMargin) * 3/4 + TopMargin;
  len = sprintf(str,"%.2f",Yextent / 4);
  x = LeftMargin - FontWidth * len - 4;
  DrawText(x,y + FontHeight / 2 - 2,str);
  DrawLine(LeftMargin - 3,y,LeftMargin + 3,y); //Tick line

  y = (BottomMargin - TopMargin) * 1/2 + TopMargin;
  len = sprintf(str,"%.2f",Yextent / 2);
  x = LeftMargin - FontWidth * len - 4;
  DrawText(x,y + FontHeight / 2 - 2,str);
  DrawLine(LeftMargin - 3,y,LeftMargin + 3,y); //Tick line

  y = (BottomMargin - TopMargin) * 1/4 + TopMargin;
  len = sprintf(str,"%.2f",Yextent * 3/4);
  x = LeftMargin - FontWidth * len - 4;
  DrawText(x,y + FontHeight / 2 - 2,str);
  DrawLine(LeftMargin - 3,y,LeftMargin + 3,y); //Tick line

  y = TopMargin - FontHeight / 2;
  len = sprintf(str,"%.2f",Yextent);
  x = LeftMargin - FontWidth * len - 4;
  vPen p2(128,0,0);
  SetPen(p2);
  myfont.SetFontValues(vfSerif,12);
  SetFont(myfont);
  DrawText(x,y + FontHeight,str);
  DrawLine(LeftMargin - 3,TopMargin,LeftMargin + 3,TopMargin);

  //Finally, graph out each dat set
  PlotLines(data);

 //Draw Intensity Mark
  if (data != 1 && Mark > 0.0) {
    myfont.SetFontValues(vfSerif,12);
    SetPen(p2);
    SetFont(myfont);
    double edef = myApp->Params.alpha0 * log(1 + myApp->Params.alpha1 *
	          (exp(myApp->Params.alpha2*Mark)-1));
    y = Height - (FontHeight+5);
    x = LeftMargin + 2;
    sprintf(str,"Coverage: %.2f        Expected Defects %.2f",Mark,edef);
    DrawText(x,y,str);

    vPen p3(128,0,0,2);
    SetPen(p3);
    x = int(LeftMargin + (Xint * Mark));
    y = BottomMargin - int(Yint * edef);
    DrawLine(x,y-9,x,y+8);
    DrawLine(x-9,y,x+8,y);
  }
}

/*****************************************************************************
  Method: PlotLines
  Parameters: which, 1 for original data, 2 for model data
  Purpose: PlotLines is called after the background for the graph has been
  drawn. It is responsible for drawing the actual graph.
*****************************************************************************/

void CovGraphPane::PlotLines(int which)
{
  double* x;
  double* y;
  int av;
  CoverageDataSet* cd;
  vColor* Colors[4];

  vBrush mybrush;
  vFont myFont(vfSerif,14);
  vPen myPen;
  myPen.SetWidth(2);
  myPen.SetStyle(vTransparent);
  SetFont(myFont);
  char *estr;
  Colors[0] = new vColor(255,0,0);
  Colors[1] = new vColor(0,125,125);
  Colors[2] = new vColor(0,0,255);
  Colors[3] = new vColor(125,125,0);
  int down = 0;

  if (which == 1) { //If we have data on each enumerable, plot that enumerable
    cd = myApp->CovData;
    x = new double[cd->Size];
    y = new double[cd->Size];
    for (int i = 0; i < cd->Size; i++)
      y[i] = cd->cov_data[i].faults;
    for (int i = 1; i < 5; i++) //For each enumerable
      if (cd->avail[i]) { //if we have data on the enumerable
	for (int j = 0; j < cd->Size; j++)
	  x[j] = cd->cov_data[j].map(i);
	switch (i) {
	case 1 : estr = "Blocks";
	  break;
	case 2 : estr = "Branches";
	  break;
	case 3 : estr = "P-Uses";
	  break;
	case 4 : estr = "C-Uses";
	  break;
	}

	myPen.SetColor(*(Colors[i-1]));
#ifdef __MSWIN__
	myPen.SetStyle(vSolid);
#endif
	SetPen(myPen);
	PlotLine(x,y,cd->Size);
 #ifdef __MSWIN__
	myPen.SetStyle(vTransparent);
	SetPen(myPen);
#endif
	mybrush.SetColor(*(Colors[i-1]));
	
	DrawText(LeftMargin+16,TopMargin+24+down,estr);
	down += 24;
      }
  }

  else if (which == 2) {
    cd = myApp->CovModData;
    for (int i = 1; i < 5; i++)  //Find what we modeled on
      if (cd->avail[i]) av = i;
    switch (av) {
    case 1 : estr = "Blocks";
      break;
    case 2 : estr = "Branches";
      break;
    case 3 : estr = "P-Uses";
      break;
    case 4 : estr = "C-Uses";
      break;
    }
    x = new double[cd->Size];
    y = new double[cd->Size];
    for (int i = 0; i < cd->Size; i++) {
      x[i] = cd->cov_data[i].map(av);
      y[i] = cd->cov_data[i].faults;    
    }
    vPen mpen(0,0,128,2);

#ifdef __UNIX__
    mpen.SetStyle(vTransparent);
#endif
    SetPen(mpen);
    PlotLine(x,y,cd->Size);
#ifdef __MSWIN__
    mpen.SetStyle(vTransparent);
    SetPen(mpen);
#endif
    DrawText(LeftMargin+16,TopMargin+24,"Model");    

    //Plot the original data along side the model
    cd = myApp->CovData;
    delete[] x; delete[] y;
    x = new double[cd->Size];
    y = new double[cd->Size];
    for (int i = 0; i < cd->Size; i++) {
      x[i] = cd->cov_data[i].map(av);
      y[i] = cd->cov_data[i].faults;    
    }
    vPen open(0,128,0,2);
#ifdef __UNIX__
    open.SetStyle(vTransparent);
#endif
    SetPen(open);
    PlotLine(x,y,cd->Size);
#ifdef __MSWIN__
    open.SetStyle(vTransparent);
    SetPen(open);
#endif
    DrawText(LeftMargin+16,TopMargin+48,estr);
  }    

  delete[] x; delete[] y;
  //vFont f2 (vfFixed);
  //SetFont(f2);
}

/****************************************************************
  Method : PlotLine
  Parameters : x, an array of x corridinates to be plotted
               y, an array of y corridinates to be plotted
	       size, the number of corridinates to be plotted
  Purpose : To draw a line along the specified corridinates.
****************************************************************/

void CovGraphPane::PlotLine(double* x, double* y, int size)
{
  int xs,ys,xe,ye;

  xs = int(LeftMargin + (x[0] * Xint));
  ys = int(BottomMargin - (y[0] * Yint));
  if (ys < TopMargin) ys = TopMargin;
  if (ys > BottomMargin) ys = BottomMargin;

  for (int cnt = 0; cnt < size - 1; cnt++) {
    xe = int(LeftMargin + (x[cnt + 1] * Xint));
    ye = int(BottomMargin - (y[cnt + 1] * Yint));
    if (xe > RightMargin) xe = RightMargin;
    if (ye < TopMargin) ye = TopMargin;
    if (ye > BottomMargin) ye = BottomMargin;
    DrawLine(xs,ys,xe,ye);
    xs = xe;
    if (xs > RightMargin) return;
    ys = ye;
  }
}

/******************************************************************
  Method : Hscroll
  Parameters : step, new position of scroll bar
  Purpose : V callback function, called to adjust intensity mark when
  vertical scroll bar is moved.
*******************************************************************/

void CovGraphPane::HScroll(int step)
{
  int top,shown;
  GetHScroll(shown,top);
  SetHScroll(shown,top + step);
  Mark  = float(top+step) / 100.0 * Xextent;
  Redraw(0,0,0,0);
}

/******************************************************************
  Method : HPage
  Parameters : shown, dummay variable for V
               top, new position of scroll bar
  Purpose : V callback function, called to adjust intensity mark when
  vertical scroll bar is dragged.
*******************************************************************/

void CovGraphPane::HPage(int shown, int top)
{
  SetHScroll(shown,top);
  Mark = float(top) / 100.0 * Xextent;
  Redraw(0,0,0,0);
}
