/*****************************************************************
  Module : covmodel.cpp
  Contents : Functions for fitting and generating coverage model
******************************************************************/
#include <math.h>
#include "datatype.h"
#include "nrutil.h"

#define NDIM 3
#define FTOL 1.0e-9
#define MYSQR(x) ((x)*(x))

//declaration of NRC code
extern int NRC_ERROR_FLAG;
void powell(float[],float**,int,float,int*,float*,float (*)(float[]));

int numpts;
double* x;
double* y;

/**************************************************************
  Function : sumerr
  Parameters : beta[], an array of model parameters
  Purpose : Helper function for NRC code, used to linear approximate
  coverage model parameters, to get a good initial guess
*******************************************************************/
float sumerr(float beta[])
{
  float temp;
  float sum = 0.0;
 
  if ((beta[2] < 0) || (beta[3] < 0))
    return 1.0e20;

  for (int cnt = 1; cnt <= numpts; cnt++) {
    temp = 1.0 + (beta[2] * (exp(beta[3] * x[cnt])) - 1.0);
    temp = beta[1] * log(temp);
    temp -= y[cnt];
    sum += MYSQR(temp);
  }
  return sum;
}

/*************************************************************************
  Function : errlin
  Parameters : beta[], an array of model parameters
  Purpose : Computes error when fitting coverage model, used by NRC code
**************************************************************************/

float errlin(float beta[])
{
  float sum = 0.0;
  for (int cnt = 1; cnt < numpts; cnt++)
    sum += MYSQR(beta[1] * x[cnt] + beta[2] - y[cnt]);
  return sum;
}

/************************************************************************
  Function : CovFit
  Parameters : cd, the base coverage data set
               i, the enumerable the model is based on
	       a0,a1,a2, the fitted parameters
  Returns : a0,a1,a2, the fitted model parmeters
  Purpose : CovFit fits a coverage model to a data set. It is essentially
  a wrapper function for NRC code. Adapted from the Naixin Li dissertation.
*************************************************************************/

void CovFit(CoverageDataSet* cd,int i,double& a0, double& a1, double& a2)
{
  a0 = a1 = a2 = 0.0; //Init to all zero's in case of premature return

  //If faults (y vals) not availble return all 0's
  if (!cd->avail[0]) return;
  if (!cd->avail[i] || !i) return;

  x = new double[cd->Size + 1];
  y = new double[cd->Size + 1];
  if (!x || !y) return;

  numpts = cd->Size;
  double temp;
  for (int cnt = 0; cnt < cd->Size; cnt++) {
    if (i == 1) temp = cd->cov_data[cnt].block;
    else if (i == 2) temp = cd->cov_data[cnt].branch;
    else if (i == 3) temp = cd->cov_data[cnt].p_use;
    else if (i == 4) temp = cd->cov_data[cnt].c_use;
    x[cnt + 1] = temp;
    y[cnt + 1] = cd->cov_data[cnt].faults;
  }

  //p is the initial starting point for powells
  float *p = vector(1,NDIM);
  p[1] = 1.0;
  p[2] = 1.0;

  //The colums of xi contains the initial directions for the search
  float** xi;
  xi = matrix(1,NDIM,1,NDIM);
  for (int i = 1; i <= NDIM; i++)
    for (int j = 1; j <= NDIM; j++)
      xi[i][j] = (i == j ? 1.0 : 0.0);

  //Values returned by powell
  int numiter;
  float fval;

  //Get inital linear estimate
  powell(p,xi,NDIM,FTOL,&numiter,&fval,errlin);

  //Setup for search
  p[1] = 1;
  p[2] = exp(p[1]);
  p[3] = p[2];
  //p[1] = 37.9;
  //p[2] = 1.65;
  //p[3] = p[2];
  powell(p,xi,NDIM,FTOL,&numiter,&fval,sumerr);

  double oval;

  do {
    oval = fval;

    //Reinit xi for next pass
    for (int i = 1; i <= NDIM; i++)
      for (int j = 1; j <= NDIM; j++)
	xi[i][j] = (i == j ? 1.0 : 0.0);
    
    powell(p,xi,NDIM,FTOL,&numiter,&fval,sumerr);
  } while ((2 * fabs(oval - fval)) > ((oval + fval) * 1.0e-6));

  //Pass back fitted parameters
  a0 = p[1];
  a1 = p[2];
  a2 = p[3];

  //Free allocated memory
  delete[] x;
  delete[] y;
  free_vector(p,1,NDIM);
  free_matrix(xi,1,NDIM,1,NDIM);
}

/*********************************************************************
  Function : CoverageModel
  Parameters : i, the enumerable the model is to be based on
               a0,a1,a2, The fitted parameters for the model
  Purpose : To create a coverage model
*********************************************************************/
CoverageDataSet* CoverageModel(int i, double a0, double a1, double a2)
{

  CoverageDataSet* nd = new CoverageDataSet(100);
  if (!nd) return NULL;

  double ecov = 0.01;
  nd->avail[0] = 1;
  nd->avail[i] = 1;

  for (int cnt = 0; cnt < 100; cnt++) {
    if (i == 1) nd->cov_data[cnt].block = ecov;
    else if (i == 2) nd->cov_data[cnt].branch = ecov;
    else if (i == 3) nd->cov_data[cnt].p_use = ecov;
    else if (i == 4) nd->cov_data[cnt].c_use = ecov;
    nd->cov_data[cnt].faults = a0 * log(1 + a1 * (exp(a2 * ecov)-1));
    ecov += 0.01;
  }
  return nd;
}

double CovMRE(CoverageDataSet* cd, int modon)
{
  int rsize = cd->Size;
  int acalc = 0;
  double finerr = cd->cov_data[cd->Size - 1].faults;
  double fincov = cd->cov_data[cd->Size - 1].map(modon);
  double mre = 0.0;
  double a0,a1,a2;
  double perr;

  for (int i = 2; i <= rsize; i++) {
    cd->Size = i;
    CovFit(cd,modon,a0,a1,a2);
    if (!NRC_ERROR_FLAG) {
      perr = a0 * log (1 + a1 * (exp(a2 * fincov) - 1));
      mre += fabs(perr - finerr)/finerr;
      acalc++;
    }
  }
    
  return mre / acalc;
}
