/*****************************************************************************
  Module : Enhance.cpp
  Contents : Function definitions for Smoothing functions
****************************************************************************/
#include <stdlib.h>
#include <math.h>
#include "datatype.h"
#include "platform.h"
#include "proto.h"
#include "robapp.h"
#include <iostream.h>

/*****************************************************************************
  Method : FixedGrouping
  Parameters : fd, a pointer to the data to be smoothed
  size, the size of groups to use for smoothing
  Returns : A pointer to the smoothed data set, or NULL on failure.
  Purpose : Performs fixed size grouping on a data set.
  Notes : This performs fixed lump grouping, assigning the median failure time
  to the group, as described in the Robust documentation. This varies from
  the original version proposed by Naixin Li.
*****************************************************************************/

FailureDataSet* FixedGrouping(FailureDataSet* data,int gsize)
{
  FailureDataSet* smth = new FailureDataSet;
  if (!smth) return NULL;
  if (!(smth->SetSize(data->Size/gsize))) return NULL;
  smth->Type = Failure_Time;

  double lt = 0.0;
  for (int dc = gsize-1, sc = 0; sc < smth->Size; dc += gsize,sc++) {
    smth->Failure_Data[sc].count = data->Failure_Data[dc].count;
    smth->Failure_Data[sc].time = 
      (data->Failure_Data[dc].time + lt) / 2.0;
    lt = data->Failure_Data[dc].time;
  }
  
  smth->Fill();
  return smth;
  
}

//This routine performs lump grouping as described in the Naixin Li
//dissertation

//FailureDataSet* FixedGrouping(FailureDataSet* fd, int size)
//{
  //Calculate the number of data points in the smoothed data set
//  int numpoints = fd->Size;
//  numpoints--;
//  numpoints /= size;
//  numpoints++;
  
  //Create the data set for the smoothed data
//  FailureDataSet* sd = new FailureDataSet;
//  if (!sd) return NULL;
//  if (!(sd->SetSize(numpoints))) return NULL;
  
//  int odcnt = 0;
//  for (int cnt = 0; cnt < numpoints; cnt++, odcnt += size) {
//    sd->Failure_Data[cnt].count = fd->Failure_Data[odcnt].count;
//    sd->Failure_Data[cnt].time = fd->Failure_Data[odcnt].time;
//  }
  
//  sd->Type = Failure_Time; //Set to time data
//  sd->Fill(); //Fill in other data fields
//  return sd;
//}

/****************************************************************************
  Method : lumping
  Parameters : fd, the failure data to be smoothed
  Returns : A pointer to the smoothed data, or NULL on failure.
  Purpose : To perform one round of lump smoothing on a data set.
  Notes : This function should never be called directly. It performs smoothing
  for LumpGrouping, which calls it for each iteration of smoothing.
  It is not visable outside this module.
***************************************************************************/

FailureDataSet* lumping(FailureDataSet* fd)
{
  int newsize = 0;
  int rising = 1;
  
  FailureDataSet *sd = new FailureDataSet;
  if (!sd) return NULL;
  if (!sd->SetSize(fd->Size)) return NULL;
  
  for (int cnt = 1; cnt < (fd->Size - 1); cnt++) {
    if(fd->Failure_Data[cnt].intensity < fd->Failure_Data[cnt + 1].intensity) {
      if (!rising) {
	newsize++;
	sd->Failure_Data[newsize].time = fd->Failure_Data[cnt].time;
	sd->Failure_Data[newsize].count = fd->Failure_Data[cnt].count;
	rising = 1;
      }
    }
    else
      rising = 0;
  }
  
  //Always keep the first and last data points
  newsize++;
  sd->Failure_Data[0].time = fd->Failure_Data[0].time;
  sd->Failure_Data[0].count = fd->Failure_Data[0].count;
  sd->Failure_Data[newsize].time = fd->Failure_Data[fd->Size - 1].time;
  sd->Failure_Data[newsize].count = fd->Failure_Data[fd->Size - 1].count;
  sd->Size = newsize + 1; //newsize was current element, now is size of array
  sd->Type = Failure_Time; //Set to time data
  sd->Fill(); //Fill in other fields
  return sd;
}

/***************************************************************************
  Method : LumpGrouping
  Parameters : fd, the failure data set to be smoothed
  times, the number of times to run the smoothing algorithm on
  the data, defaults to 2
  Returns : The set of smoothed data, or NULL on failure
  Purpose : To perform lump smoothing on a set of data.
  Notes : This is a wrapper function. It calls lumping to lump smooth each
  data set until it has smoothed the data the specified number of
  times.
***************************************************************************/

FailureDataSet* LumpGrouping(FailureDataSet* fd, int times)
{
  FailureDataSet* scratch[2];
  int cur = 0;

  scratch[0] = lumping(fd);
  if (!scratch) return NULL;

  for (int cnt = 2; cnt <= times; cnt++) {
    scratch[cur ^ 1] = lumping(scratch[cur]);
    if (!scratch[cur ^ 1]) return NULL;
    delete scratch[cur];
    cur ^= 1;
  }
  scratch[cur]->Shrink();
  return scratch[cur];
}

/***********************************************************************
  Function : ModelBias
  Parameters : data, the original data set
               model, the model created from the data set
	       cbias, the bias in the failure count (returned)
	       ibias, the bias in the failure count (returned)
  Returns : Count and intensity bias of the model
  Purpose : To compute the bias inherient in the model
************************************************************************/
void ModelBias(FailureDataSet* data, FailureDataSet* model,
		 double& cbias, double& ibias)
{
  cbias = model->Failure_Data[data->Size - 1].count -
    data->Failure_Data[data->Size - 1].count;
  ibias = model->Failure_Data[data->Size - 1].intensity -
    data->Failure_Data[data->Size - 1].intensity;
  
}

/***************************************************************************
  Function : Recalibrate
  Parameters : od, the original data set
               md, the model to be recalibrated
               cntbias, the bias in the failure count (returned)
               intenbias, the bias in the failure intensity (returned)
  Purpose : To recalibrate a SRGM model.
***************************************************************************/

void Recalibrate(FailureDataSet* od, FailureDataSet* md,
  void (*modfunc)(double, double, double, double&, double&),
  int (*fitfunc)(FailureDataSet*, double&, double&,int),double alpha,
		 int method)
{
  double b0,b1;
  double pmu,plambda;
  double cbias;
  int rsize = od->Size;

  //printf("Pts  B Orig      Bias      Mu Orig      B Cor      Mu Cor\n");

  for (int i = 2; i < rsize; i++) {
    od->Size = i;
    fitfunc(od,b0,b1,method);
    modfunc(b0,b1,od->Failure_Data[i - 1].time,pmu,plambda);
    cbias = pmu - od->Failure_Data[i - 1].count;

    //printf("%3d    %6.2f    %6.3f    %6.2f     ",i,b0,cbias,pmu);
    b0 -= cbias;    
    modfunc(b0,b1,od->Failure_Data[i].time,pmu,plambda);
    // pmu -= cbias;
    md->Failure_Data[i].count = pmu;
    //printf("%6.2f    %6.2f\n",b0,pmu);

    if (alpha > 0.0) {
      ExpFit(od,b0,b1);
      b0 -= cbias;
      b0 /= (log(alpha) / (1.0 - 1.0/alpha));
      modfunc(b0,b1,od->Failure_Data[i].time,pmu,plambda);
    }

    md->Failure_Data[i].intensity = plambda;
  }
  od->Size = rsize;
}
