/*****************************************************************************
  Module : faildata.cpp
  Contents : Method definitions for FailureDataSet
  The FailureDataSet encapsulates failure information about a data set. It
  can store data data read in from an FDS formated file, data that has been
  filtered and smoothed, or data produced by one of the software relability
  growth models.
*****************************************************************************/
#include <ctype.h>
#include <string.h>
#include <fstream.h>
#include "datatype.h"
#include "platform.h"

//To save time reading a file, a large chunk of memory is allocated.
//After the file is read, the program frees up the extra memory. However,
//If the enough memory is not allocated originally, it could cause the
//program to crash. MAX_DATA_POINTS controls how many data points can be
//read from a file before memory overwrite occurs.
#define MAX_DATA_POINTS 1024

/*****************************************************************************
  Constructor : FailureDataSet
  Parameters : None
*****************************************************************************/

FailureDataSet::FailureDataSet(void)
{
  Size = 0;
  Failure_Data = NULL;
  FileName = NULL;
  Type = Empty;
}

/***************************************************************************
  Destructor : FailureDataSet
***************************************************************************/

FailureDataSet::~FailureDataSet(void)
{
  if (Failure_Data) delete[] Failure_Data;
  delete[] FileName;
}

/***************************************************************************
  Operator : []
  Parameters : index, the element of the data set to return
  Returns : A reference to one of the members of the Failure_Data array
  Purpose : A common use of a FailureDataObject is to access the indivudual
  data points that it stores. This operator faciltates that. It is a late
  addition to the class.
**************************************************************************/
FailureDataPoint& FailureDataSet::operator[](int index)
{
  if (index < 0 || index >= Size)
    index = 0;
  return Failure_Data[index];
}


/***************************************************************************
  Method : OpenFile
  Parameters : fname, the name of the file to be opened
  Returns : 0 on failure, 1 on success
  Purpose : To read in failure data from a file
  
  Format of file : The input file must be a plain ascii text file, with two
  number to each line. The first number is an integer with the failure count.
  The second a floating point number for either the time or interval between
  failures. The first line may be a comment. If this comment is either "time"
  or "interval" then that is taken as the type of data the file holds,
  otherwise it is assumed to hold failure time data. Each line of data should
  not exceed 80 characters. This is the FDS file format described in the
  Robust manual.
*****************************************************************************/

int FailureDataSet::OpenFile(char* fname)
{
  //Reset the data members before reading in a new data set
  if (Failure_Data) delete[] Failure_Data;
  Size = 0;
  //FileName = NULL;
  Type = Empty;
  
  //Create a new stream and bind it to the file fname
  //If the open fails, return failure
  ifstream fin(fname);
  if (fin.tellg() < 0) return 0;
  FileName = new char[strlen(fname)+1];
  strcpy(FileName,fname);
  
  //Allocate a bunch of memory for the read operation
  Failure_Data = new FailureDataPoint[MAX_DATA_POINTS];
  if (!Failure_Data) return 0;
  
  //Read the first line from the file, and determine if it is data or a
  //comment. If it is a comment, attempt to determine a data type and set
  //skip = 1. If it is data, increment Size.
  char str[81];
  int skip = 0;
  
  fin.getline(str,80);
  
  if (isalpha(str[0])) { //If first line is comment...
    skip = 1;
    strlwr(str); //Convert string to all lower case for comparisons
    if (!strcmp(str,"interval"))
      Type = Failure_Interval;
    else
      Type = Failure_Time;
  }
  else
    Type = Failure_Time; //Default to failure time data
  
  if (!skip) fin.seekg(0); //If there was not a comment, go back to the beginning
  
  float temp;
  fin >> Failure_Data[Size].count;
  while (!fin.eof()) {
    fin >> temp;
    if (temp > 0.0) {
      if (Type == Failure_Interval) {
      	Failure_Data[Size].interval = temp;
	      Failure_Data[Size].time = 0.0;
      }
      else {
        Failure_Data[Size].time = temp;
        Failure_Data[Size].interval = 0.0;
      }
    }
    Size++;
    fin >> Failure_Data[Size].count;
  }

  Shrink(); //Free up additional allocated space
  Fill(); //Fill in the other data
  return 1;
}

/****************************************************************************
  Method : SaveFile
  Parameters : fname, the name of the file to be written to. Defaults to the
  name of the original data file if none is given.
  Returns : 0 on failure, 1 on success.
  Purpose : To store a failure data set in a disk file.
  Notes : Commented out because it is not used by Robust
****************************************************************************/

int FailureDataSet::SaveFile(char* fname)
{
  //Use the original file name as the default
  if (!fname) fname = FileName;
  
  //Open the file
  ofstream fout(fname);
  
  //Determine what type of data is to be written, and provide the appropriate
  //comment at the start of the file.
  if (Type == Failure_Interval)
    fout << "interval" << endl;
  else
    fout << "time" << endl;
  
  //Write the data
  for (int cnt = 0; cnt < Size; cnt++) {
    fout << Failure_Data[cnt].count << "        ";
    if (Type == Failure_Interval)
      fout << Failure_Data[cnt].interval << endl;
    else
      fout << Failure_Data[cnt].time << endl;
  }
  return 1;
}

/****************************************************************************
  Method : FillInterval
  Purpose : FillInterval fills in the time interval data based on the time
  that each failure count was taken
***************************************************************************/

void FailureDataSet::FillInterval(void)
{
  Failure_Data[0].interval = Failure_Data[0].time;
  for (int cnt = 1; cnt < Size; cnt++)
    Failure_Data[cnt].interval =
      Failure_Data[cnt].time - Failure_Data[cnt - 1].time;
}

/***************************************************************************
  Method : FillTime
  Purpose : FillTime fills in the time each failure count was taken based on
  the interval between counts.
*****************************************************************************/

void FailureDataSet::FillTime(void)
{
  Failure_Data[0].time = Failure_Data[0].interval;
  for (int cnt = 1; cnt < Size; cnt++)
    Failure_Data[cnt].time =
      Failure_Data[cnt - 1].time + Failure_Data[cnt].interval;
}

/****************************************************************************
  Method : FillIntensity
  Purpose : To compute and store the failure intensity of the data set.
*****************************************************************************/

void FailureDataSet::FillIntensity(void)
{
  Failure_Data[0].intensity = Failure_Data[0].count/Failure_Data[0].interval;
  for (int cnt = 1; cnt < Size; cnt++)
    Failure_Data[cnt].intensity =
      float(Failure_Data[cnt].count - Failure_Data[cnt - 1].count)
      / Failure_Data[cnt].interval;
  
}

/****************************************************************************
  Method : Fill
  Returns : 0 if data type (and thus the appropriate fill) is unknown,
  otherwise it returns 1.
  Purpose : Fill is a wrapper function that calls the appropriate fill for
  the failure data set, it also handels filling in the failure
  intensity.
****************************************************************************/

int FailureDataSet::Fill(void)
{
  if (Type == Empty) return 0;
  if (Type == Failure_Time) FillInterval();
  else FillTime();
  FillIntensity();
  return 1;
}

/****************************************************************************
  Method : SetSize
  Parameters : size, the new number of data points
  Returns : 0 on failure, 1 on success.
  Purpose : To change the number of data points allocated to the data set.
  Most often used by modeling routines to prepare a FailureDataSet object
  to be filled with failure data.
****************************************************************************/

int FailureDataSet::SetSize(int newsize)
{
  //Free any allocated memory
  delete Failure_Data;
  Size = 0;
  
  //Allocate memory
  Failure_Data = new FailureDataPoint[newsize];
  if (!Failure_Data) return 0;
  
  //Set the size
  Size = newsize;
  return 1;
}

/****************************************************************************
  Method : Shrink
  Purpose : To reduce the size of Failure_Data to be no larger than what the
  Size member indicates. Used by functions which over allocated Failure_Data.
****************************************************************************/

void FailureDataSet::Shrink(void)
{
  //Allocate the new data set of the appropriate size
  FailureDataPoint* nd = new FailureDataPoint[Size];
  if (!nd) return; //Failure to shrink the data set results in the existing
  //data set remaining intact
  
  //Copy the data
  for (int cnt = 0; cnt < Size; cnt++)
    nd[cnt] = Failure_Data[cnt];
  
  //Free up previously allocated memory
  delete [] Failure_Data;
  Failure_Data = nd;
}
