/**********************************************************
Module : models.cpp
Contents : Modeling and evaluation functions for the
different models.
**********************************************************/
#include <stdio.h>
#include <math.h>
#include "datatype.h"
#include "proto.h"
#include "platform.h"

//Robust uses a modified version of the NRC code. The original code
//aborts under certain conditions. The robust versions traps this error and
//passed it back through the global NRC_ERROR_FLAG
extern int NRC_ERROR_FLAG;

void mnbrak (float*,float*,float*,float*,float*,float*,float(*)(float));
float golden(float,float,float,float(*)(float),float,float*);
/***************************************************************************
  Function: XxxModel
  Parameters: b1,b1  b0,b1 are the parameters Beta0 and Beta1 for the model
              time   the time for which we want to compute the model
	      mu     the failure count at time t, as returned by the model
	      lambda the failure intensity at time t, as returned by the model
  Returns: mu and lambda
  Purpose: This is the main modeling routine, used to generate the SRG models
****************************************************************************/
//Just in case we need the constant e
//#define M_E             2.7182818284590452354

/*************************************************************************
  Function: XxxEval
  Parameters: x, the time at which the model is evaluated
              a, an array with 3 elements, a[0] is ignored, a[1] and a[2]
	         are Beta0 and Beta1 for the model
	      yfit, failure intensity at time t for model, returned
	      dyda, dyda[1] contains the deriviative with respect to Beta0,
	            dyda[2] contains the deriviative with respect to Beta1
	      ma, the size of a and dyda arrays, always 2*
  Purpose: This function is called by NRC code when doing a non-linear fit
  of the model.
***********************************************************************/

/***************************************************************************
  Function: XxxPred
  Parameters: b1,b1  b0,b1 are the parameters Beta0 and Beta1 for the model
              time   the time
	      lambda the failure intensity	
	      mu     the failure count
  Returns: two of mu, lambda, and time
  Purpose: The XxxPred functions make predictions for a model based on b0
  and b1. One of the other parameters should be none zero, the other two
  parameters are use to return their repective values for the model when the
  non-zero parameter is taken at its given value.
***************************************************************************/

/************************************************************************
  Function: XxxFit
  Parameters: fd, the failure data set to be fitted
              b0,b1, the parameters that the functions fits
  Returns: b0,b1, the fitted model parameters
  Purpose: These routines perform fitting of the model in questions. Usually
  they just to least squares fitting, but fall back to non-linear fitting if
  this fails.
**************************************************************************/

/*************************************************************************
  Model : Logarithmic
  Mu : Beta0 * ln(1 + Beta1 * Time)
  Lambda : (Beta0 * Beta1) / (1 + Beta1 * Time)
************************************************************************/

void LogModel(double b0, double b1, double time, double& mu, double& lambda)
{
  mu = b0 * log(1.0 + b1 * time);
  lambda = (b0 * b1) / (1.0 + b1 * time);
}

void LogEval(float x, float a[], float* yfit, float dyda[], int)
{
  float tmp = (1.0 + a[2] * x);
  *yfit = (a[1] * a[2]) / tmp;
  dyda[1] = a[2] / tmp;
  dyda[2] = (a[1] / tmp) - (a[1] * a[2] * x / (tmp * tmp));
}

void LogPred(double b0, double b1, double& time, double& lambda, double& mu)
{
  if (time != 0.0) {
    mu = b0 * log(1 + b1 * time);
    lambda = (b0 * b1) / (1.0 + b1 * time);
  }
  else if (lambda != 0.0) {
    time = -(lambda - (b0 * b1)) / (lambda * b1);
    mu = b0 * log(1 + b1 * time);
  }
  else if (mu != 0.0) {
    time = (exp(mu/b0) - 1) / b1;
    lambda = (b0 * b1) / (1.0 + b1 * time);
  }
}

FailureDataSet* _sme_data = NULL;

float mle_err(float b1)
{
  int n = _sme_data->Size;
  float lvalue = 0.0;
  for (int i = 0; i < n; i++)
    lvalue += 1 / (1 + b1 * _sme_data->Failure_Data[i].time);
  lvalue /= b1;

  float tn = _sme_data->Failure_Data[n-1].time;
  float tmp = 1 + b1 + tn;
  float rvalue = (n * tn) / (tmp * log(tmp));
  return fabs(lvalue-rvalue);
}

int LogMLEFit(FailureDataSet* data, double& beta0, double& beta1)
{
  float ax,bx,cx;
  float fa,fb,fc;
  float res;
  _sme_data = data;

  ax = 5.0e-4;
  bx = 4.0e-4;
  mnbrak(&ax,&bx,&cx,&fa,&fb,&fc,mle_err);
  golden (ax,bx,cx,mle_err,3e-6,&res);

  beta1 = res;
  beta0 = data->Size / (log(1 + beta1 * data->Failure_Data[data->Size - 1].time));
  return 1;
}

/**************************************************************************
  LogLSQFit(data,beta0,beta1)
  Purpose : To use least squares estimation to fit a data set to a model.
  Parameters :
      data : The data set to fit
      beta0,beta1 : The returned fit parameters
  Returns : 1 on success, 0 on failure
***************************************************************************/

int LogLSQFit(FailureDataSet* data, double& beta0, double &beta1)
{
  int size = data->Size;
  double* x = new double[size];
  double* y = new double[size];
  for (int cnt = 0; cnt < size; cnt++) {
    x[cnt] = data->Failure_Data[cnt].time;
    y[cnt] = 1/data->Failure_Data[cnt].intensity;
  }

  double slope, intercept;
  LeastSquares(x,y,size,slope,intercept);
  delete []x; delete []y;
  beta0 = 1/slope;
  beta1 = 1/(intercept * beta0);
  if (beta0 < 0.0 || beta1 < 0.0) return 0;
  return 1;
}

int LogFit(FailureDataSet* data, double& beta0, double& beta1, int method)
{
  if (method == MLE) return LogMLEFit(data,beta0,beta1);
  else return LogLSQFit(data,beta0,beta1);
}
/***********************************************************************
  Model : Exponential
  Mu : Beta0 * (1 - exp(-Beta1 * Time))
  Lambda : Beta0 * Beta1 * exp(-Beta1 * Time)
************************************************************************/

void ExpModel(double b0, double b1, double time, double& mu, double& lambda)
{
  mu = b0 * (1.0 - exp(-b1 * time));
  lambda = b0 * b1 * exp(-b1 * time);
}

void ExpEval(float x, float a[], float* yfit, float dyda[], int ma)
{
  float tmp = exp(-a[2] * x);
  *yfit = a[1] * a[2] * tmp;
  dyda[1] = a[2] * tmp;
  dyda[2] = a[1] * tmp - a[1] * a[2] * a[2] * tmp;
}

void ExpPred(double b0, double b1, double& time, double& lambda, double& mu)
{
  if (time != 0.0) {
    mu = b0 * (1.0 - exp(-b1 * time));
    lambda = b0 * b1 * exp(-b1 * time);
  }
  else if (lambda != 0.0) {
    time = log(lambda / (b0 * b1)) / -b1;
    mu = b0 * (1.0 - exp(-b1 * time));
  }
  else if (mu != 0.0) {
    time = -log((mu - b0)/-b0 ) / b1;
    lambda = b0 * b1 * exp(-b1 * time);
  }
}

FailureDataSet* _sme_data_exp = NULL;

float mle_err_exp(float b1)
{
  float err,tt,tn;
  int n = _sme_data_exp->Size;
  
  for (int i = 0; i < n; i++)
    tt = _sme_data_exp->Failure_Data[i].time;
  tn = _sme_data_exp->Failure_Data[n-1].time;

  err = (n / b1) - (n * tn) / (exp(b1 * tn) - 1) - tt;
  return fabs(err);
}

int ExpMLEFit(FailureDataSet* data, double& beta0, double& beta1)
{
  float ax,bx,cx;
  float fa,fb,fc;
  float res;
  _sme_data_exp = data;

  ax = 5.0e-4;
  bx = 4.0e-4;
  mnbrak(&ax,&bx,&cx,&fa,&fb,&fc,mle_err_exp);
  golden (ax,bx,cx,mle_err_exp,3e-6,&res);

  beta1 = res;
  beta0 = data->Size / (1 - exp(-beta1 *
				data->Failure_Data[data->Size-1].time));
  return 1;
}

/**************************************************************************
  ExpLSQFit(data,beta0,beta1)
  Purpose : To use least squares estimation to fit a data set to a model.
  Parameters :
      data : The data set to fit
      beta0,beta1 : The returned fit parameters
  Returns : 1 on success, 0 on failure
***************************************************************************/
int ExpLSQFit(FailureDataSet* data, double& beta0, double &beta1)
{
  int size = data->Size;
  double* x = new double[size];
  double* y = new double[size];
  for (int cnt = 0; cnt < size; cnt++) {
    x[cnt] = data->Failure_Data[cnt].time;
    y[cnt] = log(data->Failure_Data[cnt].intensity);
  }

  double slope, intercept;
  LeastSquares(x,y,size,slope,intercept);
  delete []x; delete []y;
  beta1 = -slope;
  beta0 = exp(intercept)/beta1;
  if (beta0 < 0.0 || beta1 < 0.0) return 0;
  return 1;
}

/*****************************************************************************
  ExpFit(data,beta0,beta1,method)
  Purpose : To fit a data set to the exponential model using the specified
  method.
  Parameters :
      data : The data set to fit
      beta0,beta1 : The returned fit parameters
      method : Either LSQ or MLE to specify the appropriate method, values
      defined in srgm.h
  Returns : 1 on success, 0 on failure      
*****************************************************************************/
int ExpFit(FailureDataSet* data, double& beta0, double& beta1, int method)
{
  if (method == MLE) return ExpMLEFit(data,beta0,beta1);
  else return ExpLSQFit(data,beta0,beta1);
}

/*************************************************************************
  Model : Power
  Mu : Beta0 * Time ^ Beta1
  Lambda : Beta0 * Beta1 * Time ^ (Beta1 - 1)
************************************************************************/

void PowModel(double b0, double b1, double time, double& mu, double& lambda)
{
  mu = b0 * pow(time,b1);
  lambda = b0 * b1 * pow(time,b1 - 1);
}

void PowEval(float x, float a[], float* yfit, float dyda[], int)
{
  float tmp = pow(x,a[2] - 1);
  *yfit = a[1] * pow(x,a[2]);
  dyda[1] = a[2] * x * exp(a[2] - 1);
  dyda[2] = a[1] * tmp + a[1] * a[2] * log(x) * tmp;
}

void PowPred(double b0, double b1, double& time, double& lambda, double& mu)
{
  if (time != 0.0) {
    mu = b0 * pow(time,b1);
    lambda = b0 * b1 * pow(time,b1 - 1);
  }
  else if (lambda != 0.0) {
    time = pow(lambda/(b0*b1),1/(b1-1));
    mu = b0 * pow(time,b1);
  }
  else if (mu != 0.0) {
    time = pow(mu/b0,1/b1);
    lambda = b0 * b1 * pow(time,b1 - 1);
  }
}
/**************************************************************************
  PowMLEFit(data,beta0,beta1)
  Purpose : To use maxium likelyhood estimation to fit a data set
  to a model.
  Parameters :
      data : The data set to fit
      beta0,beta1 : The returned fit parameters
  Returns : 1 on success, 0 on failure
***************************************************************************/
int PowMLEFit(FailureDataSet* fds, double& beta0, double& beta1)
{
  double sumt = 0.0;
  double tn = fds->Failure_Data[fds->Size-1].time;

  for (int i = 0; i < fds->Size-1; i++)
    sumt += log(tn / fds->Failure_Data[i].time);
  beta1 = fds->Size / sumt;
  beta0 = fds->Size / pow(tn,beta1);
  if (beta1 > 1.0 || beta0 < 0) return 0;
  return 1;
}

/**************************************************************************
  PowLSQFit(data,beta0,beta1)
  Purpose : To use least squares estimation to fit a data set to a model.
  Parameters :
      data : The data set to fit
      beta0,beta1 : The returned fit parameters
  Returns : 1 on success, 0 on failure
***************************************************************************/
int PowLSQFit(FailureDataSet* data, double& beta0, double &beta1)
{
  int size = data->Size;
  double* x = new double[size];
  double* y = new double[size];
  for (int cnt = 0; cnt < size; cnt++) {
    x[cnt] = log(data->Failure_Data[cnt].time);
    y[cnt] = log(data->Failure_Data[cnt].intensity);
  }

  double slope, intercept;
  LeastSquares(x,y,size,slope,intercept);
  delete []x; delete []y;
  beta1 = slope + 1.0;
  beta0 = exp(intercept)/beta1;
  if (beta0 < 0.0 || beta1 < 0.0) return 0;
  return 1;
}

/*****************************************************************************
  PowFit(data,beta0,beta1,method)
  Purpose : To fit a data set to the power model using the specified
  method.
  Parameters :
      data : The data set to fit
      beta0,beta1 : The returned fit parameters
      method : Either LSQ or MLE to specify the appropriate method, values
      defined in srgm.h
  Returns : 1 on success, 0 on failure      
*****************************************************************************/
int PowFit(FailureDataSet* data, double& beta0, double& beta1, int method)
{
  if (method == MLE) return PowMLEFit(data,beta0,beta1);
  else return PowLSQFit(data,beta0,beta1);
}

//Deleyad S model starts here
/*************************************************************************
  Model : Delayed S-Shape
  Mu : Beta0 * (1 - (1 + Beta1) * exp(-Beta1 * Time))
  Lambda : Beta0 * Beta1 ^ 2 * time * exp(-Beta1 * Time)
  Notes : Linear fitting is not correct.
************************************************************************/

void DelSModel(double b0, double b1, double time, double& mu, double& lambda)
{
  mu = b0 * (1 - (1 + b1) * exp(-b1 * time));
  lambda = b0 * b1 * b1 * time * exp(-b1 * time);
}

void DelSEval(float x, float a[], float* yfit, float dyda[], int)
{
  long double temp = EXPFUNC(-a[2] * x);
  *yfit = a[1] * a[2] * a[2] * x * temp;
  dyda[1] = a[2] * a[2] * x * temp;
  dyda[2] = (2 * a[1] * a[2] * x -
    a[1] * a[2] * a[2] * x * x) * temp;
}

void DelSPred(double b0, double b1, double& time, double& lambda, double& mu)
{
  if (time != 0.0) {
    mu = b0 * ( 1 - (1 + b1) * exp(-b1 * time));
    lambda = b0 * b1 * b1 * time * exp(-b1 * time);
  }
  else if (lambda != 0.0) {
    time = 1.0;// W(-lambda / (b0 * b1))/b1;
    mu = b0 * ( 1 - (1 + b1) * exp(-b1 * time));
  }
  else if (mu != 0.0) {
    time = -log(- (mu - b0)/(b0+b0*b1))/b1;
    lambda = b0 * b1 * b1 * time * exp(-b1 * time);
  }
}

int DelSFit(FailureDataSet* data, double& b0, double &b1, int method)
{
  int size = data->Size;
  double* x = new double[size];
  double* y = new double[size];
  for (int cnt = 0; cnt < size; cnt++) {
    x[cnt] = data->Failure_Data[cnt].time;
    y[cnt] = log(data->Failure_Data[cnt].intensity /
		 data->Failure_Data[cnt].time);
  }

  double slope, intercept;
  int didfix = -1;
  size++;
  do {
    didfix++;
    size--;
    if (size < 2) return 0;
    LeastSquares(x,y,size,slope,intercept);
    b1 = -slope;
    b0 = exp(intercept) / (b1 * b1);
  } while (b0 <= 0.0);
  delete []x; delete []y;
  NRC_ERROR_FLAG = 0;
  if (didfix ) ParmFit(data,b0,b1,DelSEval);
  return 1;
}

