/**************************************************************************
  Module : modfuncs.cpp
  Contents : Functions with are used to generate and manipulate software
  relability growth models.
***********************************************************************/

#include <math.h>
#include "nrutil.h"
#include "datatype.h"
#include "proto.h"

/***********************************************************************
  Function : Model
  Parameters : data, the failure data set the model is based on
               b0,b1, parameters for the model to be generated
	       func, one of the XxxModel functions, which returns
	             the mu and lambda values for a given time
  Return : A failure data set representing the model
  Purpose : To generate a SRGM model
*********************************************************************/
FailureDataSet* Model(FailureDataSet* data, double b0, double b1,
		      void (*func)(double,double,double,double&,double&))
{
  //Allocate the new data set
  FailureDataSet* nd = new FailureDataSet;
  if (!nd) return NULL;
  if (!nd->SetSize(data->Size)) return NULL;
  nd->Type = SRGModel;

  double time,mu,lambda;
  //For each entry in the original dat set, create one for the model
  for (int cnt = 0; cnt < nd->Size; cnt++) {
    time = data->Failure_Data[cnt].time;
    nd->Failure_Data[cnt].time = time;
    nd->Failure_Data[cnt].interval = data->Failure_Data[cnt].interval;
    func(b0,b1,time,mu,lambda);
    nd->Failure_Data[cnt].count = mu;
    nd->Failure_Data[cnt].intensity = lambda;
  }
  return nd;
}

/************************************************************************
  Function : ExpandModel
  Parameters : od, the original data set
               md, the model to be expanded
	       bo,b1, the model parameters
	       func, the model function (XxxModel family)
  Returns : md, the expanded model
  Purpose : To expand the model by 10%
  Notes : Primary used by robust in 'zoom' feature. Probably should be
  rewritten to perform recalibrate if called for, currently this is done
  by another function (checked for after model expanded).
*************************************************************************/

int ExpandModel(FailureDataSet* od, FailureDataSet*& md, double b0, double b1,
		void (*func)(double,double,double,double&,double&))
{
  FailureDataSet* md_t = md;
  if (od == NULL) od = md;

  int newelm = md->Size / 10 + 1;
  double curtime = md->Failure_Data[md->Size - 1].time;
  double timestep = (curtime * 0.1);
  timestep /= newelm;
  int os = od->Size;

  FailureDataSet* tmp = Model(od,b0,b1,func);
  FailureDataSet* nd = new FailureDataSet;

  if (!nd || !tmp) return 0;
  if (!nd->SetSize(os + newelm)) return 0;

  for (int cnt = 0; cnt < tmp->Size; cnt++)
    nd->Failure_Data[cnt] = tmp->Failure_Data[cnt];

  double mu,lambda;
  for (int cnt = tmp->Size; cnt < nd->Size; cnt++) {
    curtime += timestep;
    nd->Failure_Data[cnt].time = curtime;
    nd->Failure_Data[cnt].interval = timestep;
    func(b0,b1,curtime,mu,lambda);
    nd->Failure_Data[cnt].count = mu;
    nd->Failure_Data[cnt].intensity = lambda;
  }
  md = nd;
  delete tmp;
  delete md_t; //Can't use existing model because it might be recalibrated
  return 1;
}

/*************************************************************************
  Function : ShrinkModel
  Parameters : fd, the data set to be shrunk
  Returns : fd, the shrunken data set
  Purpose : ShrinkModel shrinks fd to 90% of its previous size.
*************************************************************************/

void ShrinkModel(FailureDataSet*& fd)
{
  int cnt = fd->Size - 1;
  float time = fd->Failure_Data[cnt].time * 0.9;

  while (fd->Failure_Data[cnt].time > time) cnt--;
  cnt++;
  if (cnt < 2) return;
  fd->Size = cnt;
  fd->Shrink();
}

/**********************************************************************
  Function : MRE
  Parameters : data, the original data set
               func, the modeling function, usually one of the XxxModel family
	       fitfunc, the fitting function for the model in question
	       smet, a set of static metrics, defaults to NULL
  Returns : The mean relative error of the model applied to the data set
  Purpose : To computer the mean relative error of the model in question.
***************************************************************************/
double MRE(FailureDataSet* data,
	   void (*func)(double,double,double,double&,double&),
	   int (*fitfunc)(FailureDataSet*,double&, double&,int),int recal,
	   StaticMetrics* smet, int (*sfunc)(StaticMetrics,double&,double&),
	   int method)
{
  double b0,b1,p;
  double err = 0.0;
  int rsize = data->Size;
  double tn = data->Failure_Data[rsize - 1].time;
  double un = data->Failure_Data[rsize - 1].count;
  double tmp;
  double cntbias;
  int accalc = 0;
  double er;
  int etype;

  for (int cnt = 2; cnt <= rsize; cnt++) {
    data->Size = cnt;
    if (!fitfunc(data,b0,b1,method)) continue;
    accalc++;
    
    if (smet && sfunc) {
      double sb0,sb1;
      //Force indirect est
      etype = smet->EstType;
      smet->EstType = 0;
      sfunc(*smet,sb0,sb1);
      smet->EstType = etype;

      if (func == LogModel) {
	b0 *= 0.75; b1 *= 0.75;
	b0 += 0.25 * sb0;
	b1 += 0.25 * sb1;
      }
      else if (func == ExpModel) {
	if (b0 > (sb0*2.0) || b0 < (sb0*0.5))
	  b0 = sb0;
      }

    }
        
    if (recal) {
      func(b0,b1,data->Failure_Data[cnt-1].time,cntbias,tmp);
      cntbias -= data->Failure_Data[cnt-1].count;
      b0 -= cntbias;
    }

    func(b0,b1,tn,p,tmp);
    err += fabs(p - un) / un;
  }
  return err / accalc;
}

double MPE(FailureDataSet* data,
	   void (*func)(double,double,double,double&,double&),
	   int (*fitfunc)(FailureDataSet*,double&, double&,int),int recal,
	   StaticMetrics* smet, int (*sfunc)(StaticMetrics,double&,double&),
	   int method)
{
  double b0,b1,p;
  double err = 0.0;
  int rsize = data->Size;
  double tn = data->Failure_Data[rsize - 1].time;
  double un = data->Failure_Data[rsize - 1].count;
  double tmp,p2;
  double cntbias;
  int accalc = 0;
  int etype;

  for (int cnt = 2; cnt <= rsize; cnt++) {
    data->Size = cnt;
    if (!fitfunc(data,b0,b1,method)) continue;
    accalc++;
    /*
      //Stabilization should not be performed for short term predictions.
    if (smet && sfunc) {
      double sb0,sb1;
      //Force indirect est
      etype = smet->EstType;
      smet->EstType = 0;
      sfunc(*smet,sb0,sb1);
      smet->EstType = etype;
      b0 *= 0.75; b1 *= 0.75;
      b0 += 0.25 * sb0;
      b1 += 0.25 * sb1;
    }
    */  
    if (recal) {
      func(b0,b1,data->Failure_Data[cnt-1].time,cntbias,tmp);
      cntbias -= data->Failure_Data[cnt-1].count;
      b0 -= cntbias;
    }

    func(b0,b1,tn,p,tmp);
    func(b0,b1,data->Failure_Data[cnt-1].time,p2,tmp); 
    tmp = un - data->Failure_Data[cnt-1].count;
    if (tmp == 0.0) continue;
    tmp = fabs(tmp - (p - p2)) / tmp;
    //printf("%9.2f %9.2f\n",data->Failure_Data[cnt-1].time / tn,tmp);
    err += tmp;
    //  printf("%9.2f %9.2f\n",data->Failure_Data[cnt-1].time / tn,err/accalc);
  }
  return err / accalc;
}

/*************************************************************
  Function : ParmFit
  Parameters : fd, the data set to fit
  b0,b1, use to return the fitted parameters
  Purpose : To fit a data set to the logarithmic model
  Notes : Adapted from the numerical recipies example book,
  this is primarly a wrapper function
************************************************************/
//This macros calls mrqmin, and aborts if it encounters an error
#define nonlin  mrqmin(x,y,sig,fd->Size,a,ia,2,covar,alpha,&chisq,evalfunc,&alamda);if (NRC_ERROR_FLAG) return -1
//Prototype for NRC code
void mrqmin(float[], float[], float[], int, float[], int[],
        int, float**, float**, float*,
        void (*funcs)(float, float [], float *, float [], int), float *);
extern int NRC_ERROR_FLAG;

int ParmFit(FailureDataSet* fd, double& b0, double& b1,
	     void (*evalfunc)(float,float[],float*,float[],int))
{
  //Set parms for nrc code
  //We just need to pass these
  float** covar = matrix(1,2,1,2);
  float** alpha = matrix(1,2,1,2);

  //This array holds the parameters being fit
  float* a = vector(1,2);
  a[1] = b0;  //These values are initial guess, they are arbitrary
  a[2] = b1; //If better guess can be found or computed, use them

  //This is an array(vector) of 0s and 1s.
  //If ia[x] is 1, it means that parm a[x] can is to be fitted
  //If ia[y] is 0, it means that parm a[y] is to be held constant
  int* ia = ivector(1,2);
  ia[1] = 1;
  ia[2] = 1;

  float alamda = -1.0; //We use this to control init of mrqmin
  float chisq; //This tells us when we are done

  //Prepare model data for nrc routines
  //sig contains standard deviations for each data point, don't know why
  float* x = vector(1,fd->Size);
  float* y = vector(1,fd->Size);
  float* sig = vector(1,fd->Size);

  for (int cnt = 0; cnt < fd->Size; cnt++) {
    x[cnt + 1] = fd->Failure_Data[cnt].time;
    y[cnt + 1] = fd->Failure_Data[cnt].intensity;
    sig[cnt + 1] = 1.0; //Roommate says this should work, I don't understand
    //the math involved
  }
  //Done setting up parameters

  //Start work here
  //Init mrqmin with alambda = -1.0

  nonlin;

  //Main loop, iterate while chisq improving
  double ochisq;
  int lastit = 0;
  int flag = 1;
  while (flag) {
    ochisq = chisq;
    nonlin;
    if (fabs(chisq - ochisq) < 0.01) lastit++;
    else lastit = 0;
    if (lastit == 2) flag = 0;
  }

  //Completion work, alamba = 0.0
  alamda = 0.0;
  nonlin;

 //Return parameters
  b0 = a[1];
  b1 = a[2];


  //Free up allocated memory
  free_matrix(covar,1,2,1,2);
  free_matrix(alpha,1,2,1,2);
  free_vector(x,1,fd->Size);
  free_vector(y,1,fd->Size);
  free_vector(sig,1,fd->Size);
  free_vector(a,1,2);
  free_ivector(ia,1,2);

  return 1;
}

/***********************************************************************
  Function : LeastSquares
  Parameters : x[], y[], a set of x,y data points to be fit
               size, the number of x,y data points to be fit
	       slope,intercept, the slope and intercept of the fitted line
  Returns : slope and intercept of the fitted line
  Purpose : To perform least squares fitting on a data set
***********************************************************************/

void LeastSquares(double x[], double y[], int size, double& slope, double& intercept)
{
  //Compute the average time
  double avgtime = 0.0;
  for (int cnt = 0; cnt < size; cnt++)
    avgtime += x[cnt];
  avgtime /= size;
  
  double sumxy = 0.0;
  double sumxx = 0.0;
  double sumy = 0.0;
  double xt;
  
  for (int cnt = 0; cnt < size; cnt++) {
    xt = x[cnt] - avgtime;
    sumxx += xt * xt;
    sumxy += xt * y[cnt];
    sumy += y[cnt];
  }

  slope = sumxy/sumxx;
  intercept = sumy/size - sumxy/sumxx * avgtime;
}

/**********************************************************************
  Function : LaplaceFactor
  Parameters : data, the failure data set to be analyized
  Returns : The laplace factor of the data set
  Purpose : To compute the laplace factor of a data set. The laplace factor
  is a measure of trend in a data set. When it is negative it represnts
  relability growth, a positive trend represent relability decrease. Values
  over or under 1.96 represent a stable trend.
*************************************************************************/

double LaplaceFactor(FailureDataSet* data)
{
  double term,time;
  double top = 0.0;
  double bot = 0.0;
 
  time = data->Failure_Data[data->Size - 1].time;

  for (int n = 0; n < (data->Size - 1); n++)
    for (int j = 0; j <= n; j++)
      top += data->Failure_Data[j].interval;
  top /= double(data->Size - 1.0);
  top -= time / 2.0;


  term = sqrt(1 / (12.0 * (data->Size - 1.0)));
  for (int j = 0; j < data->Size; j++)
    bot += data->Failure_Data[j].interval * term;

  return top/bot;
}
