/**************************************************************************
  Module : Robapp.cpp
  Contents : Class methods for robustApp, the main application class. Also,
  contains AppMain, the entry point for V GUI Toolkit programs.
  It declares the global variable myApp, the instance of the
  application class.
**************************************************************************/
#define MAINMOD
#include <stdio.h> //for sprintf
#include <math.h>
#include <v/vfilesel.h>
#include <string.h>
#include "platform.h"
#include "robapp.h"
#include "menudef.h"
#include "numsel.h"
#include "proto.h"
#include "pdiag.h"

//Used for trapping errors in the NRC code
extern int NRC_ERROR_FLAG;

/****************************************************************************
  Constructor: robustApp
****************************************************************************/
robustApp::robustApp(void) 
#ifdef __MSWIN__
  : vApp("Robust",0,750,550)
#else
  : vApp("Robust")
#endif
{
  OrgData = NULL;
  SmthData = NULL;
  ModData = NULL;
  CovData = NULL;
  CovModData = NULL;
  win1 = NULL;
  win2 = NULL;
  win3 = NULL;
  win4 = NULL;
  win5 = NULL;
  win6 = NULL;
  win7 = NULL;
  Params.data1 = 0; Params.data2 = 0;
}

/****************************************************************************
  Destructor: robustApp
***************************************************************************/
robustApp::~robustApp(void)
{
  delete vn;
  //  if (win1) win1->CloseWin();
  //if (win2) win2->CloseWin();
  //if (win3) win3->CloseWin();
  //if (win4) win4->CloseWin();
  //if (OrgData) delete OrgData;
  if (SmthData) delete SmthData;
  if (ModData) delete ModData;
  if (smdiag) delete smdiag;
  if (CovData) delete CovData;
  if (CovModData) delete CovModData;
}

/****************************************************************************
  Method : NewAppWin
  Parameters : win, the new window to add
   	       name, the name of the new window
               h, the height of the new window
               w, the width of the new window
               vAppWinInfo, a pointer to the window information structure to
               be associated with the window
   Returns : A pointer to the new window created
   Purpose : This method registered the specified window in the windowing
   	     system, placing it on the screen ready to receive commands.
	     If no window is supplied, it creates a new text window.
****************************************************************************/

vWindow *robustApp::NewAppWin(vWindow* win, char* name, int h, int w,
			      vAppWinInfo* winInfo)
{
  vWindow* mywin = win;
  vAppWinInfo* myinfo = winInfo;
  char* myname = name;
  if (myname == NULL)
    myname = "Robust";
  if (mywin == NULL)
    mywin = new txtWin(myname,h,w);
  if (myinfo == NULL)
    myinfo = new vAppWinInfo(myname);
  return vApp::NewAppWin(mywin,myname,h,w,myinfo);
}

/****************************************************************
  Method : GenModel
  Purpose : To generate a software reliability growth model.
*****************************************************************/
void robustApp::GenModel(int statmod)
{
  FailureDataSet* ds = NULL;
  int statstub = 0;
  int goodfit = 0;
  double k = 0.25;
  double sb0,sb1;

  EnsureWin(2);  //Make sure that the model data window exists
  //Determine which data set to work with
  if (SmthData) ds = SmthData;
  else if (OrgData) ds = OrgData;
  else if (statmod) {
    ds = StatStub(myApp->StatMet);
    if (!ds) Message("Out of Memory");
    statstub = 1;
    win2->SetTitle("Static Model");
  }
  else return;

  if (statmod)
    goodfit = Params.statfunc(myApp->StatMet,Params.beta0,Params.beta1);
  else {
    goodfit = Params.fitfunc(ds,Params.beta0,Params.beta1,Params.fit_type);
    if (Params.stabilize && myApp->Params.mtype == Logarithmic) {
      goodfit &= Params.statfunc(myApp->StatMet,sb0,sb1);
      Params.beta0 *= (1-k); Params.beta1 *= (1-k);
      Params.beta0 += k * sb0;
      Params.beta1 += k * sb1;
    }
    else if (Params.stabilize && myApp->Params.mtype == Exponential) {
      goodfit &= Params.statfunc(myApp->StatMet,sb0,sb1);
      if (Params.beta0 < sb0/2.0 || Params.beta0 > sb0*2.0)
	Params.beta0 = sb0;
    }
  }

  if (!goodfit) {
    Message("This data does not fit that model");
    myApp->Params.mtype = NotPresent;
    return;
  }

  if (ModData) {
    delete ModData;  //Delete any previous model data
    ModData = NULL;
  }
  ModData = Model(ds,Params.beta0,Params.beta1,Params.modfunc);

  if (!ModData) {
    Message("Out of Memory while creating model!");
    theApp->Exit();
  }

  //Recalibrate if neccessary
  if (Params.recalibrate) {
    double tmp = 0.0;
    if (Params.mtype == Logarithmic) tmp = StatMet.Alpha;
    Recalibrate(ds,ModData,Params.modfunc,Params.fitfunc,tmp,Params.fit_type);
    Params.recalibrated = 1;
  }
  else {
    Params.recalibrated = 0;
    if (!statstub) ModelBias(ds,ModData,Params.cntbias,Params.intenbias);
    else Params.cntbias = Params.intenbias = 0.0;
  }

  if (statstub) delete ds;

  //Display the model
  win2->SetDataSet(ModelData);
  Params.data2 = ModelData;
  if (win3) win3->SetDataSet(Params.data2,Params.data1);
  if (win7) win7->SetDataSet(Params.data2,Params.data1);
  if (win6) win6->Update();
  MenuEntryState(m_ShowNum,1);
  MenuEntryState(m_Graph,1);
}

/****************************************************************
  Method : GenCovModel
  Purpose : To generate a software reliability growth model based on
  coverage.
*****************************************************************/
void robustApp::GenCovModel(int i)
{
  if (!CovData) return;
  if (!CovData->avail[i]) return;
  //Make sure that the model data window exists
  EnsureWin(2);

  //Delete any previous model data
  if (CovModData) delete CovModData;
  
  CovFit(CovData,i,Params.alpha0,Params.alpha1,Params.alpha2);
  CovModData = CoverageModel(i,Params.alpha0,Params.alpha1,Params.alpha2);

  if (!CovModData) {
    Message("Out of Memory while creating model!");
    theApp->Exit();
  }

  win2->SetDataSet(CoverModelData);
  Params.data2 = CoverModelData;
  if (win4) win4->SetDataSet(2);
  if (win6) win6->Update();
  MenuEntryState(m_ShowNum,1);
}

/****************************************************************************
  Method : Smooth
  Parameters : type, 1 for fixed smoothing, 2 for lump smoothing
  Purpose : Wrapper function to handel smoothing
**************************************************************************/

void robustApp::Smooth(int type)
{
    if (!OrgData) return;
    EnsureWin(1);

    intSelDiag *dg;
    char* stitle;
    char* msg;
    int def,low,high,slevel;

    if (type == 1) {
      stitle = "Group Smoothing";
      msg = "Group Size: ";
      def = OrgData->Size/10;
      high = OrgData->Size;
      low = 2;
    }
    else {
      stitle = "Lump Smoothing";
      msg = "Times to Lump: ";
      def = 2;
      high = 5;
      low = 1;
    }      

    dg = new intSelDiag(stitle,low,high,def);
    slevel = dg->GetNum(msg);
    if (SmthData) delete SmthData;

    if (type == 1) SmthData = FixedGrouping(OrgData,slevel);
    else SmthData = LumpGrouping(OrgData,slevel);

    if (!SmthData) {
      Message("Out of Memory!");
      theApp->Exit();
    }
    win1->SetDataSet(SmoothData);
    Params.data1 = SmoothData;
    if (win3) {
      if (ModData) win3->SetDataSet(Params.data2,Params.data1);
      else win3->SetDataSet(Params.data1,0);
    }
    delete dg;
}

/****************************************************************************
   Method : About
   Purpose : To display the about robust dialog box.
****************************************************************************/

void robustApp::About(void)
{
   about->ShowDialog("");
}

/**********************************************************
  Method : SetTitle
  Parameters : fname, the name of the data file read in
  Purpose : To parse the complete path name of a new data,
  and set the title parameters to match the file name.
***********************************************************/

void robustApp::SetTitle(char* fname)
{
  char delim;

#ifdef __UNIX__
  delim = '/';
#endif
#ifdef __MSWIN__
  delim = '\\';
#endif

  int last = -1;
  int cnt = 0;
  
  while (fname[cnt] != '\0') {
    if (fname[cnt] == delim) last = cnt;
    cnt++;
  }

  last++;
  cnt = 0;
  while (fname[last] != '\0' && cnt < 15) {
    Params.title[cnt] = fname[last];
    cnt++; last++;
  }
  Params.title[cnt] = '\0';
}

/****************************************************************************
  Method: OpenFailureDataSet
  Returns: 1 on success, 0 on failure
  Purpose: Open failure data allows the user to select the file to be opened,
   	   and reads the data in, displaying it in the window.
****************************************************************************/

int robustApp::OpenFailureDataSet(char* fn)
{
  EnsureWin(1);
  
  win1->data = 0;
  printf("creating dialog\n");
  vFileSelect fdiag(this);
  
  int ftype = 0;
  char* flist[3];
  char fname[128];
  flist[0] = "*.fds";
  flist[1] = "*.cds";
  flist[2] = NULL;
  fname[0] = '\0';

  if (!fn) {
    printf("opening\n");
    if (!fdiag.FileSelect("Open File",fname,128,flist,ftype))
      return 1; //Return success if the user cancels the operation
  }
  else
    strcpy(fname,fn);

  if (OrgData) delete OrgData;

  if (fname[strlen(fname) - 3] == 'c') ftype = 2;
  else ftype = 1;

  if (ftype == 1) { //failure data
    OrgData = new FailureDataSet;
    if (!OrgData->OpenFile(fname)) return 0;
    SetTitle(fname); 
    win1->SetDataSet(OrigData);
    Params.data1 = OrigData;
    FailMenus();
  }
  else if (ftype == 2) { //coverage data
    CovData = new CoverageDataSet;
    if (!CovData->OpenFile(fname)) return 0;
    SetTitle(fname); 
    win1->SetDataSet(CoverData);
    Params.data1 = CoverData;
    OrgData = NULL;
    CovMenus();
  }

  MenuEntryState(m_ShowNum,0);
  if (win2) win2->Clear();
  if (win3) CloseAppWin(win3);
  if (win4) CloseAppWin(win4);
  if (win6) CloseAppWin(win6);
  if (win7) CloseAppWin(win7);
  if (SmthData) delete SmthData;
  if (ModData) delete ModData;
  if (CovModData) delete CovModData;
  SmthData = NULL;
  ModData = NULL;
  CovModData = NULL;
  Params.data2 = 0;
  Params.recalibrated = 0;
  Params.recalibrate = 0;
  Params.stabilize = 0;
  Params.mtype = NotPresent;
  return 1;
}

void robustApp::Revert(void)
{
  if (OrgData) { //failure data
    SetTitle(OrgData->FileName); 
    win1->SetDataSet(OrigData);
    Params.data1 = OrigData;
    FailMenus();
  }
  else if (CovData) { //coverage data
    SetTitle(CovData->FileName); 
    win1->SetDataSet(CoverData);
    Params.data1 = CoverData;
    OrgData = NULL;
    CovMenus();
  }
  else return;

  MenuEntryState(m_ShowNum,0);
  if (win2) win2->Clear();
  if (win3) CloseAppWin(win3);
  if (win7) CloseAppWin(win7);
  if (win4) CloseAppWin(win4);
  if (SmthData) delete SmthData;
  if (ModData) delete ModData;
  if (CovModData) delete CovModData;
  SmthData = NULL;
  ModData = NULL;
  CovModData = NULL;
  Params.data2 = 0;
  Params.recalibrated = 0;
  Params.recalibrate = 0;
  Params.stabilize = 0;
  Params.mtype = NotPresent;  
}

void robustApp::Reset(void)
{
  Params.data1 = 0;
  if (win1) win1->Clear();
  if (CovData) delete CovData;
  if (OrgData) delete OrgData;
  StartMenus();
  if (win2) win2->Clear();
  if (win3) CloseAppWin(win3);
  if (win7) CloseAppWin(win7);
  if (win4) CloseAppWin(win4);
  if (SmthData) delete SmthData;
  if (ModData) delete ModData;
  if (CovModData) delete CovModData;
  OrgData = NULL;
  SmthData = NULL;
  ModData = NULL;
  CovData = NULL;
  CovModData = NULL;
  Params.data2 = 0;
  Params.recalibrated = 0;
  Params.recalibrate = 0;
  Params.stabilize = 0;
  Params.mtype = NotPresent;  
}

void robustApp::SaveDataSet(int ds)
{
  int idx;
  if (ds == SmoothData) idx = 0;
  else if (ds == ModelData) idx = 1;
  else return;

  char fname[64];
  char* flist[2] = {"*.fds",NULL};
  char* dname[2] = {"Smoothed Data File Name","Model File Name"};
  int fi = 0;
  vFileSelect fselect(theApp);

  if(!fselect.FileSelect(dname[idx],fname,64,flist,fi)) return;

  if (idx == 0) SmthData->SaveFile(fname);
  else ModData->SaveFile(fname);
}

/***********************************************************************
  Method : CloseNotify
  Parameters : win, the window being closed
  Purpose : CloseNotify is called when one of robusts windows is closed.
  This enables robust to keep track of which windows it has open.
*******************************************************************/

void robustApp::CloseNotify(vCmdWindow* win)
{
  if (win1 == win) win1 = NULL;
  else if (win2 == win) win2 = NULL;
  else if (win3 == win) win3 = NULL;
  else if (win4 == win) win4 = NULL;
  else if (win5 == win) win5 = NULL;
  else if (win6 == win) win6 = NULL;
  else if (win7 == win) win7 = NULL;

}

void robustApp::EnsureWin(int num)
{
  char title[32];

  switch (num) {
  case 1 :
    if (!win1) {
      win1 = new txtWin("Original Data",txth,txtw);
      NewAppWin(win1,"Original Data",txth,txtw,NULL);
      win1->SetValue(m_Recal,Params.recalibrate,Checked);
      //win2->SetValue(m_Stable,Params.stabilize,Checked);
      win1->SetValue(m_MLE,Params.fit_type,Checked);
      //win2->SetValue(m_MLE,Params.fit_type,Checked);
      win1->SetValue(m_LSQ,!Params.fit_type,Checked);
      //win2->SetValue(m_LSQ,!Params.fit_type,Checked);


#ifdef __UNIX__
      win1->SetPosition(100,100);
#endif
    }
    break;
  case 2 :
    if (!win2) {
      win2 = new txtWin("Model Data",txth,txtw);
      NewAppWin(win2,"Model Data",txth,txtw,NULL);
      win2->SetValue(m_Recal,Params.recalibrate,Checked);
      win2->SetValue(m_Stable,Params.stabilize,Checked);
      win2->SetValue(m_LSQ,!Params.fit_type,Checked);
      win2->SetValue(m_MLE,Params.fit_type,Checked);
#ifdef __MSWIN__
      win2->SetPosition(0,txth+25);
#endif
      
#ifdef __UNIX__
      win2->SetPosition(100,txth+178);
#endif
    }
    break;
  case 3 :
    if (!win3) {
      win3 = new grphWin();
      NewAppWin(win3,"Graph Display",375,450,NULL);
      win3->SetValue(m_Recal,Params.recalibrate,Checked);
      win3->SetValue(m_Stable,Params.stabilize,Checked);
      win3->SetPosition(300,100);
      win3->SetValue(m_LSQ,!Params.fit_type,Checked);
      win3->SetValue(m_MLE,Params.fit_type,Checked); 
   }
    break;
  case 4 :
    if (!win4) {
      win4 = new covgrphWin();
      NewAppWin(win4,"Graph Display",400,500,NULL);
      win4->SetValue(m_Recal,Params.recalibrate,Checked);
      win4->SetValue(m_Stable,Params.stabilize,Checked);
      //win4->SetPosition(300,100);
    }
    break;

  case 5 :
    if (!win5) {
      win5 = new helpWin("Help Topics",txth,txtw);
      NewAppWin(win5,"Help Topics",txth,txtw,NULL);
      win5->SetPosition(200,200);
    }
    break;
  case 6 :
    if (!win6) {
      win6 = new infoWin("Model Information",200,200);
      NewAppWin(win6,"Model Information",200,200,NULL);
      win6->SetPosition(200,200);
    }
    break;
  case 7 :
    if (!win7) {
      win7 = new grphMuWin();
      NewAppWin(win7,"Graph Display",375,450,NULL);
      win7->SetValue(m_Recal,Params.recalibrate,Checked);
      win7->SetValue(m_Stable,Params.stabilize,Checked);
      win7->SetPosition(300,(txth+178));
      win7->SetValue(m_LSQ,!Params.fit_type,Checked);
      win7->SetValue(m_MLE,Params.fit_type,Checked);
    }
    break;
  }
}


/************************************************************************
  Method : ShowNumbers
  Purpose : ShowNumbers displays the model information dialog box.
************************************************************************/

void robustApp::ShowNumbers(void)
{
  EnsureWin(6);
  win6->Update();
}

/*****************************************************************************
  Method : Predict
  Parameters : predon, an integer telling what the prediction is being made on
  1 = predict based on time
  2 = predict based on failure intensity
  3 = predict based on failure count
  Purpose : To display the predict on, and results of prediction dialog boxes
***************************************************************************/

void robustApp::Predict(int predon)
{
  if (!ModData && !CovModData) return;

  PredDialog* pdiag;
  Pred2Dialog* pdiag2;
  double value;
  int tmp;

  //Do nothing if no model
  if (!ModData) return;
  
  pdiag = new PredDialog(predon);
  pdiag->ShowModalDialog(NULL,tmp);
  value = pdiag->value;
  delete pdiag;
  
  pdiag2 = new Pred2Dialog(predon,value);
  pdiag2->ShowModalDialog(NULL,tmp);
  delete pdiag2;
}

/****************************************************************************
  Method : MenuEntryState
  Parameters : id, the menu item to change
               val, 1 to make sensitive, 0 to make insensitive
  Purpose : To change the sensitivity of a menu item in all open windows.
**************************************************************************/

void robustApp::MenuEntryState(ItemVal id,ItemVal val)
{
  if (win1) win1->SetValue(id,val,Sensitive);
  if (win2) win2->SetValue(id,val,Sensitive);
  if (win3) win3->SetValue(id,val,Sensitive);
  if (win4) win4->SetValue(id,val,Sensitive);
  if (win7) win1->SetValue(id,val,Sensitive);

}

/*******************************************************
  Method : CovMenus
  Purpose : To adjust the sensitivty of menu items as
  appropriate for working with coverage data sets.
*********************************************************/

void robustApp::CovMenus(void)
{
  MenuEntryState(m_CMBL,1);
  MenuEntryState(m_CMBR,1);
  MenuEntryState(m_CMPU,1);
  MenuEntryState(m_CMCU,1);
  
  MenuEntryState(m_Graph,1);
  MenuEntryState(m_Fixed,0);
  MenuEntryState(m_Lump,0);
  MenuEntryState(m_Recal,0);
  MenuEntryState(m_Stable,0);
  MenuEntryState(m_MLE,0);
  MenuEntryState(m_LSQ,0);

  MenuEntryState(m_LogModel,0);
  MenuEntryState(m_ExpModel,0);
  MenuEntryState(m_PowModel,0);
  MenuEntryState(m_DlSModel,0);
  MenuEntryState(m_SmoothMenu,0);
  MenuEntryState(m_PTime,0);
  MenuEntryState(m_PInten,0);
  MenuEntryState(m_PCount,0);

}

/*******************************************************
  Method : FailMenus
  Purpose : To adjust the sensitivty of menu items as
  appropriate for working with failure data sets.
*********************************************************/

void robustApp::FailMenus(void)
{
  MenuEntryState(m_CMBL,0);
  MenuEntryState(m_CMBR,0);
  MenuEntryState(m_CMPU,0);
  MenuEntryState(m_CMCU,0);

  MenuEntryState(m_Graph,1);
  MenuEntryState(m_Fixed,1);
  MenuEntryState(m_Lump,1);
  MenuEntryState(m_Recal,1);
  MenuEntryState(m_Stable,1);
  MenuEntryState(m_MLE,1);
  MenuEntryState(m_LSQ,1);



  MenuEntryState(m_SmoothMenu,1);
  MenuEntryState(m_PTime,1);
  MenuEntryState(m_PInten,1);
  MenuEntryState(m_PCount,1);

  int sw = 0;
  if (Params.data1) sw = 1;
  MenuEntryState(m_LogModel,sw);
  MenuEntryState(m_ExpModel,sw);
  MenuEntryState(m_PowModel,sw);
  MenuEntryState(m_DlSModel,sw);
}

void robustApp::StartMenus(void)
{
  MenuEntryState(m_CMBL,0);
  MenuEntryState(m_CMBR,0);
  MenuEntryState(m_CMPU,0);
  MenuEntryState(m_CMCU,0);
  
  MenuEntryState(m_Fixed,0);
  MenuEntryState(m_Lump,0);
  MenuEntryState(m_Recal,0);
  MenuEntryState(m_Stable,0);
  MenuEntryState(m_MLE,0);
  MenuEntryState(m_LSQ,0);

  MenuEntryState(m_LogModel,0);
  MenuEntryState(m_ExpModel,0);
  MenuEntryState(m_PowModel,0);
  MenuEntryState(m_DlSModel,0);
  MenuEntryState(m_SmoothMenu,0);
  MenuEntryState(m_Graph,0);
  MenuEntryState(m_ShowNum,0);

  MenuEntryState(m_PTime,0);
  MenuEntryState(m_PInten,0);
  MenuEntryState(m_PCount,0);
}

//Single instance of the robust application class
static robustApp* rApp = new robustApp;

/***************************************************************************
  Method : AppMain
  Parameters : argc, the number of arguments
  argv, pointers to strings containing command line parameters
  Purpose : This is the entry point for the program, it simply creates the
  first window.
****************************************************************************/

int AppMain(int argc, char** argv)
{
   myApp->win1 = new txtWin("Original Data",txth,txtw);
   myApp->win2 = new txtWin("Model Data",txth,txtw);
   myApp->NewAppWin(myApp->win1,"Original Data",txth,txtw,NULL);
   myApp->NewAppWin(myApp->win2,"Model Data",txth,txtw,NULL);
   myApp->win1->Clear();
   myApp->win2->Clear();
   
   #ifdef __MSWIN__
      myApp->win2->SetPosition(0,txth+25);
   #endif

   #ifdef __UNIX__
      myApp->win1->SetPosition(100,50);
      myApp->win2->SetPosition(100,txth+128);
   #endif

   myApp->win1->RaiseWindow();
   myApp->StartMenus();

   myApp->win1->SetValue(m_MLE,1,Checked);
   myApp->win2->SetValue(m_MLE,1,Checked);
   // myApp->win2->SetValue(m_LSQ,0,Checked);

   myApp->vn = new vNoticeDialog(theApp);
   myApp->smdiag = new StatMetDiag(theApp);
   myApp->hdiag = new helpDialog(theApp);
   myApp->about = new AboutBox(theApp);

   myApp->Params.beta0 = 0.0;
   myApp->Params.beta1 = 0.0;
   myApp->Params.cntbias = 0.0;
   myApp->Params.intenbias = 0.0;
   myApp->Params.mre = 0.0;
   myApp->Params.recalibrated = 0;
   myApp->Params.recalibrate = 0;
   myApp->Params.stabilize = 0;
   myApp->Params.mtype = NotPresent;
   myApp->Params.title[0] = '\0';
   myApp->Params.fit_type = MLE;
   
   if (argc > 1) {
     if (!myApp->OpenFailureDataSet(argv[1])) {
       Message("Could not open file");
       return 0;
     }
     if (argc > 2) {
       ItemVal mt;
       if (!strcmp(argv[2],"log")) mt = m_LogModel;
       else if (!strcmp(argv[2],"exp")) mt = m_ExpModel;
       else if (!strcmp(argv[2],"pow")) mt = m_PowModel;
       else if (!strcmp(argv[2],"block")) mt = m_CMBL;
       else if (!strcmp(argv[2],"branch")) mt = m_CMBR;
       else if (!strcmp(argv[2],"puse")) mt = m_CMPU;
       else if (!strcmp(argv[2],"cuse")) mt = m_CMCU;
       myApp->AppCommand(NULL,mt,0,C_Menu);
     }
   }

   return 0;
}


