/**********************************************************************
  Module : statmod.cpp
  Contents : Fitting and support functions for static models.
*************************************************************************/

#include <math.h>
#include "datatype.h"

/***************************************************************
  Function : StatExpParms
  Parameters : sm, a set of static metrics
               b0,b1 the parameters determined by the function
  Returns : b0,b1 parameters statically estimated
  Purpose : To estimate static parameters for the exponential model

*****************************************************************/
int StatExpParms(StaticMetrics sm,double& b0, double& b1)
{
  float dd = sm.DefectDensity();
  b0 = 1.05 * dd * sm.Size;
  b1 = (1.2e-6 / dd) * exp(0.05 * dd); 
  b1 /= (sm.Size * sm.Instructions / (sm.CPUrate * 1000));

  if (b0 >= 0.0 && b1 >= 0.0) return 1;
  return 0;
}

/***************************************************************
  Function : StatLogParms
  Parameters : sm, a set of static metrics
               b0,b1 the parameters determined by the function
  Returns : b0,b1 parameters statically estimated
  Purpose : To estimate static parameters for the exponential model

*****************************************************************/
int StatLogParms(StaticMetrics sm,double& b0, double& b1)
{
  if (!sm.EstType) { //if not direct estimation
    double b0e,b1e;
    double tmp1,tmp2;

    StatExpParms(sm,b0e,b1e);
    b1 = b1e * ((sm.Alpha - 1) / log(sm.Alpha));
    tmp1 = 1 / (1 - (1/sm.Alpha));
    tmp2 = b1 / b1e * log(sm.Alpha) + 1;
    tmp1 *= log(tmp2);
    b0 = b0e / tmp1;
  }
  else { // direct estimation
    double dmin;
    double fac = 1.5e-7 / 2.7182818;

    dmin = sm.DefectDensity();
    dmin = (dmin < 10) ? 2.0 : dmin / 3.0;

    fac *= (sm.CPUrate * 1000) / (sm.Instructions * sm.Size);
    b1 = fac * exp(sm.DefectDensity()/dmin);
    b0 = sm.Size * dmin;
  } 
  if (b0 >= 0.0 && b1 >= 0.0) return 1;
  return 0;
}

/********************************************************************
  Function : StatStub
  Parameters : sm, a set of static metrics
  Returns : A failure data set with time intervals out to the desired time
  Purpose : Creating a static model requires a base of existing data.
  In the event that estimation is done before testing begins, this function
  creates a data set extending to the requested time.
*********************************************************************/

FailureDataSet* StatStub(StaticMetrics sm)
{
  FailureDataSet* fd = new FailureDataSet;
  fd->SetSize(100);

  for (int cnt = 0; cnt < 100; cnt++) {
    fd->Failure_Data[cnt].count = cnt;
    fd->Failure_Data[cnt].time = sm.Time/100 * (cnt + 1);
    fd->Failure_Data[cnt].interval = sm.Time/100;
  }
  return fd;
}








