/*****************************************************************************
  Module : Txtwin.cpp
  Contents : Method definitions for the txtWin class, which displays a text
  view of a failure data set.
*****************************************************************************/

#include <stdio.h>
#include "robapp.h"
#include "menus.h"

/****************************************************************************
  Constructor: txtWin
  Parameters: name, the name of the window
  h, the height of the window
  w, the width of the window
*****************************************************************************/

txtWin::txtWin(char* name, int h, int w)
  : vCmdWindow(name,w,h)
{
  data = 0;
  mp = new vMenuPane(MainMenu);
  AddPane(mp);
  tp = new txtPane;
  AddPane(tp);
  
  ShowWindow();
  tp->ShowVScroll(1); //Show the scroll bar
  tp->SetVScroll(5,0);
  //Set font
  vFont myfont(vfFixed,12);
  tp->SetFont(myfont);
}

/****************************************************************************
  Destructor: txtWin
***************************************************************************/

txtWin::~txtWin(void)
{
  delete mp;
  delete tp;
}

/***********************************************************************
  Method : CloseWin
  Purpose : To notify robust that a window has been closed, and that
  the pointer to it is no longer valid.
*******************************************************************/

void txtWin::CloseWin(void)
{
  myApp->CloseNotify(this);
  vCmdWindow::CloseWin();
}

/******************************************************************
  Method : Clear
  Purpose : To clear a text window of any printed text.
********************************************************************/

void txtWin::Clear(void)
{
  if (data == ModelData)
    SetTitle("Model Data");
  else if (data == OrigData)
    SetTitle("Original Data");

  data = 0;
  tp->shown = 5;
  tp->Clear();
  tp->SetVScroll(5,0);
}

/*****************************************************************************
  Method: SetDataSet
  Parameters: fd, the new data set to be used.
  Purpose: SetDataSet sets the data set to be displayed by the window.
*****************************************************************************/

void txtWin::SetDataSet(int ds)
{
  data = ds;
  tp->data = ds;
  int shown = 5;
  float size;
  char title[40];

  FailureDataSet* fd;
  if (ds == OrigData) {
    sprintf(title,"Original Data - %s",myApp->Params.title);
    fd = myApp->OrgData;
    size = float(fd->Size);
  }

  else if (ds == SmoothData) {
    sprintf(title,"Smoothed Data - %s",myApp->Params.title);
    fd = myApp->SmthData;
    size = float(fd->Size);
  }

  else if (ds == ModelData) {
    sprintf(title,"Model Data - %s",myApp->Params.title);
    fd = myApp->ModData;
    size = float(fd->Size);
  }

  else if (ds == CoverData) {
    sprintf(title,"Coverage Data - %s",myApp->Params.title);
    size = myApp->CovData->Size;
  }

  else if (ds == CoverModelData) {
    sprintf(title,"Coverage Model Data - %s",myApp->Params.title);
    size = myApp->CovModData->Size;
  }

  shown = int((tp->GetRows() - 5.0)/size * 100);
  if (shown > 100) shown = 100;
  tp->shown = shown;
  tp->SetVScroll(shown,0);
  tp->ShowDataSet(0);
  SetTitle(title);
}

/****************************************************************************
  Method: ResizeText
  Parameters: rows, the new number of rows on the display
  cols, the new number of columns on the display
****************************************************************************/

void txtPane::ResizeText(int, int)
{
  SetVScroll(shown,0);
  ShowDataSet(0);
}

/****************************************************************************
  Method: ShowDataSet
  Parameters: top, the percentage through the data set to being display at
  Purpose: ShowDataSet places the failure data into the text buffer so that
  the parent class can display it. This is the code the determines
  what is shown on the screen.
*****************************************************************************/

void txtPane::ShowDataSet(int top)
{
  if (top < 0) top = 0;
  
  char buf[128];
  FailureDataSet* fd = NULL;
  CoverageDataSet* cd = NULL;
  int size = 0;

  if (data == OrigData) fd = myApp->OrgData;
  else if (data == SmoothData) fd = myApp->SmthData;
  else if (data == ModelData) fd = myApp->ModData;
  else if (data == CoverData) {
    cd = myApp->CovData;
    size = cd->Size;
  }
  else if (data == CoverModelData) {
    cd = myApp->CovModData;
    size = cd->Size;
  }

  if (!fd && !cd) return;
  if (fd) size = fd->Size;

  int rt = GetRows() - 5;
  int start = int((top / 100.0) * (size));
  int end = start + rt;
  if (end >= size) { 
    end = size - 1;
    start = end - rt;
    if (start < 0) start = 0;
  }
  Clear();

  if (data < CoverData) {
    DrawText("Failure       Failure          Failure          Failure\n");
    DrawText(" Count         Time           Interval         Intensity\n\n");
    
    for (int cnt = start; cnt <= end; cnt++) {
      sprintf(buf,"%5.1f  %15.7f  %15.7f %15.7f\n",
	      fd->Failure_Data[cnt].count, fd->Failure_Data[cnt].time,
	      fd->Failure_Data[cnt].interval, fd->Failure_Data[cnt].intensity);
      DrawText(buf);
    }
  }

  else if (data >= CoverData) {
    char buf2[128];

    strcpy(buf,"");
    if (cd->avail[0]) {
      sprintf(buf2,"Cumlative\n Faults     ");
      strcat(buf,buf2);
    }
    if (cd->avail[1]) {
      sprintf(buf2,"Blocks      ");
      strcat(buf,buf2);
    }
    if (cd->avail[2]) {
      sprintf(buf2,"Branches      ");
      strcat(buf,buf2);
    } 
    if (cd->avail[3]) {
      sprintf(buf2,"P_uses      ");
      strcat(buf,buf2);
    }
    if (cd->avail[4]) {
      sprintf(buf2,"C_uses      ");
      strcat(buf,buf2);
    }
    sprintf(buf2,"\n\n");
    strcat(buf,buf2);
    DrawText(buf);

    for (int cnt = start; cnt <= end; cnt++) {
    strcpy(buf,"");
      if (cd->avail[0]) {
	sprintf(buf2," %4.1f       ",cd->cov_data[cnt].faults);
	strcat(buf,buf2);
      }
      if (cd->avail[1]) {
	sprintf(buf2,"%4.2f       ",cd->cov_data[cnt].block*100);
	strcat(buf,buf2);
      }
      if (cd->avail[2]) {
	sprintf(buf2,"  %4.2f       ",cd->cov_data[cnt].branch*100);
	strcat(buf,buf2);
      }
      if (cd->avail[3]) {
	sprintf(buf2,"%4.2f       ",cd->cov_data[cnt].p_use*100);
	strcat(buf,buf2);
      }
      if (cd->avail[4]) {
	sprintf(buf2,"%4.2f       ",cd->cov_data[cnt].c_use*100);
	strcat(buf,buf2);
      }
      sprintf(buf2,"\n");
      strcat(buf,buf2);
      DrawText(buf);
    }
  }
}

/****************************************************************************
  Method: VScroll
  Parameters: step, the amount of adjust the scroll by
  Purpose: VScroll responsed to small step movements of the scroll bar.
*****************************************************************************/

void txtPane::VScroll(int step)
{
  int top;
  GetVScroll(shown,top);
  SetVScroll(shown,top + step);
  ShowDataSet(top + step);
}

/****************************************************************************
  Method: VPage
  Parameters: shown, the percentage of the data set being displayed
  top, the pertange of the way through the data to start display
  Purpose: VPage responds to vertical page command from the scroll bar.
*****************************************************************************/

void txtPane::VPage(int s, int top)
{
  SetVScroll(s,top);
  ShowDataSet(top);
}



